"""Tests for the contacts.YAMLEditable class"""

import unittest

from .helpers import TestYAMLEditable


class ToYamlConversion(unittest.TestCase):
    def test_yaml_quoted_special_characters(self):
        yaml_editable = TestYAMLEditable()
        yaml_editable.supported_private_objects = ["Twitter"]
        yaml_repr = """
Formatted name: Test vCard
First name: Khard
Private       :
    Twitter: \"@khard\"
"""
        yaml_editable.update(yaml_repr)
        yaml_dump = yaml_editable.to_yaml()
        self.assertIn("'@khard'", yaml_dump)

    def test_dumping_multiple_home_addresses_to_yaml(self):
        yaml_editable = TestYAMLEditable()
        yaml_editable._add_post_address(
            "home", "", "", "street 1", "zip1", "city1", "", ""
        )
        yaml_editable._add_post_address(
            "home", "", "", "street 2", "zip2", "city2", "", ""
        )
        yaml_dump = yaml_editable.to_yaml()
        self.assertIn("zip1", yaml_dump)
        self.assertIn("zip2", yaml_dump)

    def test_dumping_multiple_home_phone_number_to_yaml(self):
        yaml_editable = TestYAMLEditable()
        yaml_editable._add_phone_number("home", "1234567890")
        yaml_editable._add_phone_number("home", "0987654321")
        yaml_dump = yaml_editable.to_yaml()
        self.assertIn("1234567890", yaml_dump)
        self.assertIn("0987654321", yaml_dump)

    def test_dumping_multiple_home_email_addresses_to_yaml(self):
        yaml_editable = TestYAMLEditable()
        yaml_editable.add_email("home", "home1@example.org")
        yaml_editable.add_email("home", "home2@example.org")
        yaml_dump = yaml_editable.to_yaml()
        self.assertIn("home1", yaml_dump)
        self.assertIn("home2", yaml_dump)

    def test_empty_kind_is_included_in_yaml_format(self):
        contact = TestYAMLEditable()
        yaml = contact.to_yaml()
        self.assertIn("Kind:", yaml)

    def test_kind_is_included_in_yaml_format(self):
        contact = TestYAMLEditable(version="4.0", kind="org")
        yaml = contact.to_yaml()
        self.assertIn("Kind: org", yaml)


class ExceptionHandling(unittest.TestCase):
    def test_duplicate_key_errors_are_translated_to_value_errors(self):
        ye = TestYAMLEditable()
        with self.assertRaises(ValueError):
            ye.update("{key: value, key: again}")

    def test_parser_error_is_translated_to_value_error(self):
        ye = TestYAMLEditable()
        with self.assertRaises(ValueError):
            ye.update("{[invalid yaml")


class PrivateObjects(unittest.TestCase):
    def test_can_add_strings(self) -> None:
        ye = TestYAMLEditable()
        ye.supported_private_objects = ["foo"]
        ye._add_private_object("foo", "bar")
        self.assertEqual(ye._get_private_objects(), {"foo": ["bar"]})

    def test_can_add_several_strings_under_the_same_label(self) -> None:
        ye = TestYAMLEditable()
        ye.supported_private_objects = ["foo"]
        ye._add_private_object("foo", "bar")
        ye._add_private_object("foo", "baz")
        self.assertEqual(ye._get_private_objects(), {"foo": ["bar", "baz"]})

    def test_unsupported_private_objects_can_be_added_but_not_retrieved(self) -> None:
        ye = TestYAMLEditable()
        ye.supported_private_objects = ["foo"]
        ye._add_private_object("foo", "bar")
        ye._add_private_object("bar", "foo")
        self.assertEqual(ye._get_private_objects(), {"foo": ["bar"]})
        self.assertIn("X-BAR:foo", ye.vcard.serialize())

    def test_private_objects_can_have_an_ablabel(self) -> None:
        ye = TestYAMLEditable()
        ye.supported_private_objects = ["foo"]
        foo = ye.vcard.add("X-FOO")
        foo.value = "bar"
        foo.group = "1"
        label = ye.vcard.add("X-ABLABEL")
        label.value = "baz"
        label.group = "1"
        result = ye._get_private_objects()
        self.assertEqual(result, {"foo": [{"baz": "bar"}]})

    def test_private_objects_with_ablabels_are_sorted_by_ablabel(self) -> None:
        ye = TestYAMLEditable()
        ye.supported_private_objects = ["foo"]
        foo1 = ye.vcard.add("X-FOO")
        foo1.value = "AAA"
        foo1.group = "1"
        label1 = ye.vcard.add("X-ABLABEL")
        label1.value = "yyy"
        label1.group = "1"
        foo2 = ye.vcard.add("X-FOO")
        foo2.value = "BBB"
        foo2.group = "2"
        label2 = ye.vcard.add("X-ABLABEL")
        label2.value = "xxx"
        label2.group = "2"
        result = ye._get_private_objects()
        self.assertEqual(result, {"foo": [{"xxx": "BBB"}, {"yyy": "AAA"}]})



"""Unittests for the khard module"""

import locale
import unittest
import itertools
from argparse import Namespace
from email.headerregistry import Address
from unittest import mock

from khard import config, khard, query
from khard.contacts import Contact
from khard.khard import find_email_addresses

from .helpers import TestContact, TmpAbook, load_contact, mock_locale


class TestSearchQueryPreparation(unittest.TestCase):
    foo = query.TermQuery("foo")
    bar = query.TermQuery("bar")

    def setUp(self):
        # Set the uninitialized global variable in the khard module to make it
        # mockable. See https://stackoverflow.com/questions/61193676
        khard.config = mock.Mock()

    @staticmethod
    def _make_abook(name):
        abook = mock.Mock()
        abook.name = name
        return abook

    @classmethod
    def _run(cls, **kwargs):
        with mock.patch(
            "khard.khard.config.abooks",
            [cls._make_abook(name) for name in ["foo", "bar", "baz"]],
        ):
            return khard.prepare_search_queries(Namespace(**kwargs))

    def test_queries_for_the_same_address_book_are_joind_by_disjunction(self):
        expected = self.foo | self.bar
        prepared = self._run(
            addressbook=["foo"],
            target_addressbook=["foo"],
            source_search_terms=self.foo,
            target_contact=self.bar,
        )
        self.assertEqual(expected, prepared["foo"])

    def test_no_search_terms_result_in_any_queries(self):
        expected = query.AnyQuery()
        prepared = self._run(
            addressbook=["foo"],
            target_addressbook=["foo"],
            source_search_terms=query.AnyQuery(),
            target_contact=query.AnyQuery(),
        )
        self.assertEqual(expected, prepared["foo"])


class TestFindEmailAddress(unittest.TestCase):
    def test_empty_text_finds_none(self):
        text = ""
        addrs = find_email_addresses(text, ["from"])
        self.assertEqual([], addrs)

    def test_single_header_finds_one_address(self):
        text = """From: John Doe <jdoe@machine.example>"""
        addrs = find_email_addresses(text, ["from"])
        expected = [
            Address(display_name="John Doe", username="jdoe", domain="machine.example")
        ]
        self.assertEqual(expected, addrs)

    def test_single_header_finds_multiple_addresses(self):
        text = """From: John Doe <jdoe@machine.example>, \
                Mary Smith <mary@example.net>"""
        addrs = find_email_addresses(text, ["from"])
        expected = [
            Address(display_name="John Doe", username="jdoe", domain="machine.example"),
            Address(display_name="Mary Smith", username="mary", domain="example.net"),
        ]
        self.assertEqual(expected, addrs)

    def test_non_address_header_finds_none(self):
        text = (
            "From: John Doe <jdoe@machine.example>, "
            "Mary Smith <mary@example.net>\nOther: test"
        )
        addrs = find_email_addresses(text, ["other"])
        expected = []
        self.assertEqual(expected, addrs)

    def test_multiple_headers_finds_some(self):
        text = (
            "From: John Doe <jdoe@machine.example>, "
            "Mary Smith <mary@example.net>\nOther: test"
        )
        addrs = find_email_addresses(text, ["other", "from"])
        expected = [
            Address(display_name="John Doe", username="jdoe", domain="machine.example"),
            Address(display_name="Mary Smith", username="mary", domain="example.net"),
        ]
        self.assertEqual(expected, addrs)

    def test_multiple_headers_finds_all(self):
        text = (
            "From: John Doe <jdoe@machine.example>, "
            "Mary Smith <mary@example.net>\n"
            "To: Michael Jones <mjones@machine.example>"
        )
        addrs = find_email_addresses(text, ["to", "FrOm"])
        expected = [
            Address(
                display_name="Michael Jones",
                username="mjones",
                domain="machine.example",
            ),
            Address(display_name="John Doe", username="jdoe", domain="machine.example"),
            Address(display_name="Mary Smith", username="mary", domain="example.net"),
        ]
        self.assertEqual(expected, addrs)

    def test_finds_all_emails(self):
        text = (
            "From: John Doe <jdoe@machine.example>, "
            "Mary Smith <mary@example.net>\n"
            "To: Michael Jones <mjones@machine.example>"
        )
        addrs = find_email_addresses(text, ["all"])
        expected = [
            Address(display_name="John Doe", username="jdoe", domain="machine.example"),
            Address(display_name="Mary Smith", username="mary", domain="example.net"),
            Address(
                display_name="Michael Jones",
                username="mjones",
                domain="machine.example",
            ),
        ]
        self.assertEqual(expected, addrs)

    def test_finds_all_emails_with_other_headers_too(self):
        text = (
            "From: John Doe <jdoe@machine.example>, "
            "Mary Smith <mary@example.net>\n"
            "To: Michael Jones <mjones@machine.example>"
        )
        addrs = find_email_addresses(text, ["other", "all", "from"])
        expected = [
            Address(display_name="John Doe", username="jdoe", domain="machine.example"),
            Address(display_name="Mary Smith", username="mary", domain="example.net"),
            Address(
                display_name="Michael Jones",
                username="mjones",
                domain="machine.example",
            ),
        ]
        self.assertEqual(expected, addrs)


class TestGetContactListByUserSelection(unittest.TestCase):
    def setUp(self):
        """initialize the global config object with a mock"""
        khard.config = mock.Mock(spec=config.Config)
        khard.config.group_by_addressbook = False
        khard.config.reverse = False
        khard.config.sort = "last_name"

    def tearDown(self):
        del khard.config

    def test_uid_query_without_strict_search(self):
        q = query.FieldQuery("uid", "testuid1")
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            l = khard.get_contact_list(abook, q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, "testuid1")

    def test_name_query_with_uid_text_and_strict_search(self):
        q = query.NameQuery("testuid1")
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            l = khard.get_contact_list(abook, q)
        self.assertEqual(len(l), 0)

    def test_name_query_with_uid_text_and_without_strict_search(self):
        q = query.NameQuery("testuid1")
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            l = khard.get_contact_list(abook, q)
        self.assertEqual(len(l), 0)

    def test_term_query_without_strict_search(self):
        q = query.TermQuery("testuid1")
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            l = khard.get_contact_list(abook, q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, "testuid1")

    def test_term_query_with_strict_search_matching(self):
        q = query.TermQuery("second contact")
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            l = khard.get_contact_list(abook, q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, "testuid1")


class TestSortContacts(unittest.TestCase):
    contact1 = load_contact("contact1.vcf")
    contact2 = load_contact("contact2.vcf")
    nickname = load_contact("nickname.vcf")
    no_nickname = load_contact("no-nickname.vcf")

    def _test(self, *contacts: Contact, **kwargs):
        """Run the sort_contacts function and assert the result

        The contacts are expected to come out in the order that they are given
        in.
        """
        contacts_list = list(contacts)
        for order in itertools.permutations(contacts):
            actual = khard.sort_contacts(order, **kwargs)
            self.assertEqual(actual, contacts_list)

    def test_sorts_by_first_name_by_default(self):
        self._test(self.nickname, self.no_nickname)

    def test_reverses_sort_order(self):
        self._test(self.no_nickname, self.nickname, reverse=True)

    def test_sorting_of_korean_names(self):
        korean_c = load_contact("korean-c.vcf")
        korean_j = load_contact("korean-j.vcf")
        with mock_locale(locale.LC_COLLATE, "korean"):
            self._test(korean_j, korean_c)

    def test_can_sort_by_last_name(self):
        self._test(self.no_nickname, self.nickname, sort="last_name")

    def test_can_sort_by_formatted_name(self):
        self._test(self.contact1, self.contact2, sort="formatted_name")

    def test_group_by_addressbook(self):
        with TmpAbook(["contact1.vcf", "category.vcf"], name="one") as abook1:
            with TmpAbook(["contact2.vcf", "labels.vcf"], name="two") as abook2:
                contact1 = next(abook1.search(query.FieldQuery("uid", "testuid1")))
                category = next(abook1.search(query.NameQuery("category")))
                contact2 = next(abook2.search(query.FieldQuery("uid", "testuid2")))
                labels = next(abook2.search(query.NameQuery("labeled guy")))
        self._test(category, contact1, labels, contact2, group=True)

    def test_sort_order_for_accentuated_names(self):
        # reported in issue #127
        albert = TestContact(fn="Albert")
        eleanor = TestContact(fn="Eleanor")
        eugene = TestContact(fn="Eugene")
        zakari = TestContact(fn="Zakari")
        eric = TestContact(fn="Éric")
        with mock_locale(locale.LC_COLLATE, "fr"):
            self._test(albert, eleanor, eric, eugene, zakari, sort="formatted_name")



"""Helper code for type annotations and runtime type conversion."""

from datetime import datetime
from typing import overload


# some type aliases
Date = str | datetime
StrList = str | list[str]
PostAddress = dict[str, str]


# A default year for datetimes without one.  Python does not support datetime
# objects without a year from 3.15 onwards but vCard does.  We add a default
# year to the python object and detect it again when formatting the object.
# 1900 was the internally used default year of python's datetime object when it
# still did support instances without a year (before python 3.5).
DEFAULT_YEAR = 1900


@overload
def convert_to_vcard(name: str, value: StrList, constraint: type[str]) -> str: ...
@overload
def convert_to_vcard(name: str, value: StrList, constraint: type[list]) -> list[str]: ...
@overload
def convert_to_vcard(name: str, value: StrList, constraint: None) -> StrList: ...
def convert_to_vcard(name: str, value: StrList, constraint: None | type[str] | type[list]) -> StrList:
    """converts user input into vCard compatible data structures

    :param name: object name, only required for error messages
    :param value: user input
    :param constraint: set the accepted return type for vCard attribute
    :returns: cleaned user input, ready for vCard or a ValueError
    """
    if isinstance(value, str):
        if constraint is list:
            return [value.strip()]
        return value.strip()
    if isinstance(value, list):
        if constraint is str:
            raise ValueError(f"{name} must contain a string.")
        if not all(isinstance(entry, str) for entry in value):
            raise ValueError(f"{name} must not contain a nested list")
        # filter out empty list items and strip leading and trailing space
        return [x.strip() for x in value if x.strip()]
    if constraint is str:
        raise ValueError(f"{name} must be a string.")
    if constraint is list:
        raise ValueError(f"{name} must be a list with strings.")
    raise ValueError(f"{name} must be a string or a list with strings.")


def list_to_string(input: str | list, delimiter: str) -> str:
    """converts list to string recursively so that nested lists are supported

    :param input: a list of strings and lists of strings (and so on recursive)
    :param delimiter: the delimiter to use when joining the items
    :returns: the recursively joined list
    """
    if isinstance(input, list):
        return delimiter.join(
            list_to_string(item, delimiter) for item in input)
    return input


def string_to_list(input: str | list[str], delimiter: str) -> list[str]:
    if isinstance(input, list):
        return input
    return [x.strip() for x in input.split(delimiter)]


def string_to_date(string: str) -> datetime:
    """Convert a date string into a date object.

    :param string: the date string to parse
    :returns: the parsed datetime object
    """

    # Attempt to parse the string as any of the date and time formats supported
    # by Khard, as defined by the vCard and ISO 8601:2000 specifications.
    # Strings which define a day-of-month but not a year are ambiguous, require
    # special handling, and will be unsupported in Python >= 3.15. (They were
    # already removed in ISO 8601:2004, but remain a part of vCard.)

    # Ambiguous cases of a date with no year (--%m%d and --%m-%d).
    try:
        if string.startswith("--"):
            tmp = str(DEFAULT_YEAR) + string[2:]
            if "-" in string[2:]:
                return datetime.strptime(tmp, "%Y%m-%d")
            else:
                return datetime.strptime(tmp, "%Y%m%d")
    except ValueError:
        pass

    # Fully qualified date and time formats.
    for fmt in (
        "%Y%m%d",
        "%Y-%m-%d",
        "%Y%m%dT%H%M%S",
        "%Y-%m-%dT%H:%M:%S",
        "%Y%m%dT%H%M%SZ",
        "%Y-%m-%dT%H:%M:%SZ",
    ):
        try:
            return datetime.strptime(string, fmt)
        except ValueError:
            continue

    # Timezone formats which may contain a problematic colon.
    for fmt in ("%Y%m%dT%H%M%S%z", "%Y-%m-%dT%H:%M:%S%z"):
        try:
            return datetime.strptime("".join(string.rsplit(":", 1)), fmt)
        except ValueError:
            continue

    # All formats tried. Date cannot be parsed.
    raise ValueError



"""Tests for the vcard formatting functions"""

import unittest

from vobject.vcard import Name

from khard.address_book import VdirAddressBook
from khard.contacts import Contact
from khard.formatter import Formatter

from .helpers import vCard


class FormatLabeledField(unittest.TestCase):
    def test_labels_are_selected_alphabetically_if_no_preferred_given(self):
        labeled_field = {"some": ["thing"], "other": ["thing"]}
        preferred = []
        expected = "other: thing"
        self.assertEqual(
            expected, Formatter.format_labeled_field(labeled_field, preferred)
        )

    def test_labels_are_selected_alphabetically_if_no_preferred_matches(self):
        labeled_field = {"some": ["thing"], "other": ["thing"]}
        preferred = ["nonexistent"]
        expected = "other: thing"
        self.assertEqual(
            expected, Formatter.format_labeled_field(labeled_field, preferred)
        )

    def test_preferred_labels_are_used(self):
        labeled_field = {"some": ["thing"], "other": ["thing"]}
        preferred = ["some"]
        expected = "some: thing"
        self.assertEqual(
            expected, Formatter.format_labeled_field(labeled_field, preferred)
        )

    def test_alphabetically_first_value_is_used(self):
        labeled_field = {"some": ["thing", "more"]}
        preferred = []
        expected = "some: more"
        self.assertEqual(
            expected, Formatter.format_labeled_field(labeled_field, preferred)
        )

    def test_not_only_first_char_of_label_is_used(self):
        preferred = []
        labeled_field = {"x-foo": ["foo"], "x-bar": ["bar"]}
        expected = "x-bar: bar"
        self.assertEqual(
            expected, Formatter.format_labeled_field(labeled_field, preferred)
        )
        expected = "bar: bar"
        labeled_field = {"foo": ["foo"], "bar": ["bar"]}
        self.assertEqual(
            expected, Formatter.format_labeled_field(labeled_field, preferred)
        )


class GetSpecialField(unittest.TestCase):
    _name = Name(
        family="Family",
        given="Given",
        additional="Additional",
        prefix="Prefix",
        suffix="Suffix",
    )
    _vcard = Contact(
        vCard(fn="Formatted Name", n=_name, nickname="Nickname"),
        VdirAddressBook("test", "/tmp"),
        "",
    )

    def _test_name(self, fmt, nick, parsable, expected):
        f = Formatter(fmt, [], [], nick, parsable)
        actual = f.get_special_field(self._vcard, "name")
        self.assertEqual(expected, actual)

    def test_name_formatted_as_first_name_last_name(self):
        self._test_name(Formatter.FIRST, False, False, "Given Additional Family")

    def test_name_formatted_as_first_name_last_name_with_nickname(self):
        self._test_name(
            Formatter.FIRST, True, False, "Given Additional Family (Nickname: Nickname)"
        )

    def test_name_formatted_as_last_name_first_name(self):
        self._test_name(Formatter.LAST, False, False, "Family, Given Additional")

    def test_name_formatted_as_last_name_first_name_with_nickname(self):
        self._test_name(
            Formatter.LAST, True, False, "Family, Given Additional (Nickname: Nickname)"
        )

    def test_name_formatted_as_formatted_name(self):
        self._test_name(Formatter.FORMAT, False, False, "Formatted Name")

    def test_name_formatted_as_formatted_name_with_nickname(self):
        self._test_name(
            Formatter.FORMAT, True, False, "Formatted Name (Nickname: Nickname)"
        )

    def test_parsable_overrides_nickname_with_first_formatting(self):
        self._test_name(Formatter.FIRST, True, True, "Given Additional Family")

    def test_parsable_overrides_nickname_with_last_formatting(self):
        self._test_name(Formatter.LAST, True, True, "Family, Given Additional")

    def test_parsable_overrides_nickname_with_formatted_name(self):
        self._test_name(Formatter.FORMAT, True, True, "Formatted Name")



"""Tests for the Contact class from the contacts module."""

# pylint: disable=missing-docstring

import base64
import datetime
import unittest
from unittest import mock

from khard.contacts import Contact, multi_property_key


class ContactFormatDateObject(unittest.TestCase):
    def test_format_date_object_will_not_touch_strings(self):
        expected = "untouched string"
        actual = Contact._format_date_object(expected, False)
        self.assertEqual(actual, expected)

    def test_format_date_object_with_simple_date_object(self):
        d = datetime.datetime(2018, 2, 13)
        actual = Contact._format_date_object(d, False)
        self.assertEqual(actual, "2018-02-13")

    def test_format_date_object_with_simple_datetime_object(self):
        d = datetime.datetime(2018, 2, 13, 0, 38, 31)
        with mock.patch("time.timezone", -7200):
            actual = Contact._format_date_object(d, False)
        self.assertEqual(actual, "2018-02-13T00:38:31+02:00")

    def test_format_date_object_with_date_1900(self):
        d = datetime.datetime(1900, 2, 13)
        actual = Contact._format_date_object(d, False)
        self.assertEqual(actual, "--02-13")


class AltIds(unittest.TestCase):
    def test_altids_are_read(self):
        card = Contact.from_file(None, "test/fixture/vcards/altid.vcf")
        expected = "one representation"
        self.assertEqual(expected, card.get_first_name_last_name())


class Photo(unittest.TestCase):
    """Tests related to the PHOTO property of vCards"""

    PNG_HEADER = b"\x89PNG\r\n\x1a\n"

    def test_parsing_base64_ecoded_photo_vcard_v3(self):
        c = Contact.from_file(None, "test/fixture/vcards/photov3.vcf")
        self.assertEqual(c.vcard.photo.value[:8], self.PNG_HEADER)

    def test_parsing_base64_ecoded_photo_vcard_v4(self):
        c = Contact.from_file(None, "test/fixture/vcards/photov4.vcf")
        uri_stuff, data = c.vcard.photo.value.split(",")
        self.assertEqual(uri_stuff, "data:image/png;base64")
        data = base64.decodebytes(data.encode())
        self.assertEqual(data[:8], self.PNG_HEADER)


class MultiPropertyKey(unittest.TestCase):
    """Test for the multi_property_key helper function"""

    def test_strings_are_in_the_first_sort_group(self) -> None:
        group, _key = multi_property_key("some string")
        self.assertEqual(group, 0)

    def test_dicts_are_in_the_second_sort_group(self) -> None:
        group, _key = multi_property_key({"some": "dict"})
        self.assertEqual(group, 1)

    def test_strings_are_their_own_keys(self) -> None:
        _group, key = multi_property_key("some string")
        self.assertEqual(key, "some string")

    def test_dicts_are_keyed_by_the_first_key(self) -> None:
        _group, key = multi_property_key({"some": "dict", "more": "stuff"})
        self.assertEqual(key, "some")

    def test_all_strings_are_sorted_before_dicts(self) -> None:
        my_list = ["a", {"c": "d"}, "e", {"f": "g"}]
        my_list.sort(key=multi_property_key)  # type: ignore[arg-type]
        self.assertEqual(my_list, ["a", "e", {"c": "d"}, {"f": "g"}])



"""Module to make it possible to import the test folder as a python package and
hence make it possible to run all unittests from the top level directory with

        python -m unittest [discover]
"""



"""Tests for the config module."""

# pylint: disable=missing-docstring

import io
import logging
import os.path
from pathlib import Path
import tempfile
import unittest
import unittest.mock as mock

import configobj

from khard import config
from khard.exceptions import ConfigError


class LoadingConfigFile(unittest.TestCase):
    def test_load_non_existing_file_fails(self):
        filename = "I hope this file never exists"
        with self.assertRaises(OSError) as cm:
            config.Config._load_config_file(filename)
        self.assertTrue(str(cm.exception).startswith("Config file not found:"))

    def test_uses_khard_config_environment_variable(self):
        filename = "this is some very random string"
        with mock.patch.dict("os.environ", clear=True, KHARD_CONFIG=filename):
            with mock.patch("configobj.ConfigObj", dict):
                ret = config.Config._load_config_file("")
        self.assertEqual(ret["infile"], filename)

    def test_uses_xdg_config_home_environment_variable(self):
        prefix = "this is some very random string"
        with mock.patch.dict("os.environ", clear=True, XDG_CONFIG_HOME=prefix):
            with mock.patch("configobj.ConfigObj", dict):
                ret = config.Config._load_config_file("")
        expected = os.path.join(prefix, "khard", "khard.conf")
        self.assertEqual(ret["infile"], expected)

    def test_uses_config_dir_if_environment_unset(self):
        prefix = "this is some very random string"
        with mock.patch.dict("os.environ", clear=True, HOME=prefix):
            with mock.patch("configobj.ConfigObj", dict):
                ret = config.Config._load_config_file("")
        expected = os.path.join(prefix, ".config", "khard", "khard.conf")
        self.assertEqual(ret["infile"], expected)

    def test_load_empty_file_fails(self):
        with tempfile.NamedTemporaryFile() as name:
            with self.assertLogs(level=logging.ERROR):
                with self.assertRaises(ConfigError):
                    config.Config(name)

    @mock.patch.dict("os.environ", EDITOR="editor", MERGE_EDITOR="meditor")
    def test_load_minimal_file_by_name(self):
        cfg = config.Config("test/fixture/minimal.conf")
        self.assertEqual(cfg.editor, ["editor"])
        self.assertEqual(cfg.merge_editor, "meditor")


class DiscoverAddressBooks(unittest.TestCase):
    def test_discover_books(self):
        filename = "test/fixture/discover.conf"
        cfg = config.Config(filename)
        cfg.init_address_books()
        expected = {"broken.abook", "nick.abook", "test.abook", "minimal.abook"}
        self.assertEqual(set(cfg.abooks._abooks.keys()), expected)

    @staticmethod
    def config_file(abooks: dict) -> config.Config:
        string = "[addressbooks]\n"
        for key, value in abooks.items():
            string += f"""[[{key}]]
            path = {value["path"]}
            type = {value["type"]}
            """
        c = config.Config(io.StringIO(string))
        c.init_address_books()
        return c

    def test_simple_vdir_and_discover_abooks(self):
        c = self.config_file({
            "normal addressbook": {"path": "doc", "type": "vdir"},
            "discover addressbook": {"path": "doc/so*", "type": "discover"},
        })
        self.assertEqual(c.abooks[0].name, "normal addressbook")
        self.assertEqual(c.abooks[0].path, "doc")
        self.assertEqual(c.abooks[1].name, "source")
        self.assertEqual(c.abooks[1].path, "doc/source")

    def test_discover_abooks_ignore_files_when_globbing(self):
        c = self.config_file({
            "discover": {"path": "doc/source/*", "type": "discover"},
        })
        self.assertEqual(len(c.abooks), 2)

    def test_names_of_discover_abooks_are_the_basenames_of_the_leaf_directory(self):
        with tempfile.TemporaryDirectory() as tmp:
            tmp = Path(tmp)
            (tmp / "a" / "b" / "c").mkdir(parents=True)
            (tmp / "a" / "x" / "x1").mkdir(parents=True)
            (tmp / "a" / "x" / "x2").mkdir(parents=True)
            (tmp / "a" / "y" / "y1").mkdir(parents=True)
            (tmp / "a" / "y" / "y2").mkdir(parents=True)
            c = self.config_file({
                "discover": {"path": f"{tmp}/**", "type": "discover"}
            })
            actual = sorted([a.name for a in c.abooks])
        expected = ["c", "x1", "x2", "y1", "y2"]
        self.assertEqual(actual, expected)


class ConfigPreferredVcardVersion(unittest.TestCase):
    def test_default_value_is_3(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.preferred_vcard_version, "3.0")

    def test_set_preferred_version(self):
        c = config.Config("test/fixture/minimal.conf")
        c.preferred_vcard_version = "11"
        self.assertEqual(c.preferred_vcard_version, "11")


class Defaults(unittest.TestCase):
    def test_debug_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.debug)

    def test_default_action_defaults_to_none(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertIsNone(c.default_action)

    def test_reverse_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.reverse)

    def test_group_by_addressbook_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.group_by_addressbook)

    def test_show_nicknames_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.show_nicknames)

    def test_show_uids_defaults_to_true(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertTrue(c.show_uids)

    def test_show_kinds_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.show_kinds)

    def test_sort_defaults_to_first_name(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.sort, "first_name")

    def test_display_defaults_to_first_name(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.display, "first_name")

    def test_localize_dates_defaults_to_true(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertTrue(c.localize_dates)

    def test_preferred_phone_number_type_defaults_to_pref(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertListEqual(c.preferred_phone_number_type, ["pref"])

    def test_preferred_email_address_type_defaults_to_pref(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertListEqual(c.preferred_email_address_type, ["pref"])

    def test_private_objects_defaults_to_empty(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertListEqual(c.private_objects, [])

    def test_search_in_source_files_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.search_in_source_files)

    def test_skip_unparsable_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.skip_unparsable)

    def test_preferred_version_defaults_to_3(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.preferred_vcard_version, "3.0")

    @mock.patch.dict("os.environ", clear=True)
    def test_editor_defaults_to_vim(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.editor, ["vim"])

    @mock.patch.dict("os.environ", clear=True)
    def test_merge_editor_defaults_to_vimdiff(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.merge_editor, "vimdiff")


class Validation(unittest.TestCase):
    @staticmethod
    def _template(section, key, value):
        configspec = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), "khard", "data", "config.spec"
        )
        c = configobj.ConfigObj(configspec=configspec)
        c["general"] = {}
        c["vcard"] = {}
        c["contact table"] = {}
        c["addressbooks"] = {"test": {"path": "/tmp"}}
        c[section][key] = value
        return c

    def test_rejects_invalid_default_actions(self):
        action = "this is not a valid action"
        conf = self._template("general", "default_action", action)
        with self.assertLogs(level=logging.ERROR):
            with self.assertRaises(ConfigError):
                config.Config._validate(conf)

    def test_rejects_unparsable_editor_commands(self):
        editor = 'editor --option "unparsable because quotes are missing'
        conf = self._template("general", "editor", editor)
        with self.assertLogs(level=logging.ERROR):
            with self.assertRaises(ConfigError):
                config.Config._validate(conf)

    def test_rejects_private_objects_with_strange_chars(self):
        obj = "X-VCÄRD-EXTENSIÖN"
        conf = self._template("vcard", "private_objects", obj)
        with self.assertLogs(level=logging.ERROR):
            with self.assertRaises(ConfigError):
                config.Config._validate(conf)

    def test_rejects_private_objects_starting_with_minus(self):
        obj = "-INVALID-"
        conf = self._template("vcard", "private_objects", obj)
        with self.assertLogs(level=logging.ERROR):
            with self.assertRaises(ConfigError):
                config.Config._validate(conf)


if __name__ == "__main__":
    unittest.main()



"""Tests for the user interaction functions."""

import unittest
from unittest import mock

from khard.helpers import interactive

from .helpers import mock_stream


class Ask(unittest.TestCase):
    def test_accepts_on_of_the_given_options(self):
        with mock.patch("builtins.input", lambda _: "foo"):
            actual = interactive.ask("message", ["foo", "bar"])
        self.assertEqual("foo", actual)

    def test_does_not_accept_answers_not_in_choices(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["baz", "foo"])):
            actual = interactive.ask("message", ["foo", "bar"])
        self.assertEqual("foo", actual)

    def test_default_is_accepted_on_empty_input(self):
        with mock.patch("builtins.input", lambda _: ""):
            actual = interactive.ask("message", ["foo", "bar"], "baz")
        self.assertEqual("baz", actual)

    def test_accepts_prefix_match(self):
        with mock.patch("builtins.input", lambda _: "f"):
            actual = interactive.ask("message", ["foo", "bar"])
        self.assertEqual("foo", actual)

    def test_only_accepts_unique_prefix_match(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["ba", "bar"])):
            with mock_stream() as stdout:
                actual = interactive.ask("message", ["baz", "bar"])
        stdout = stdout.getvalue()
        self.assertEqual("bar", actual)
        self.assertIn("not specific enough", stdout)


class Select(unittest.TestCase):
    def _test(self, include_none=None):
        input_list = ["a", "b", "c"]
        if include_none is None:
            return interactive.select(input_list)
        else:
            return interactive.select(input_list, include_none)

    def test_selection_index_is_1_based(self):
        with mock.patch("builtins.input", lambda _: "1"):
            actual = self._test()
        self.assertEqual(actual, "a")

    def test_typing_a_non_number_prints_a_message_and_repeats(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["foo", "2"])):
            with mock_stream() as stdout:
                actual = self._test()
        stdout = stdout.getvalue()
        self.assertEqual(
            stdout, "Please enter an index value between 1 and 3 " "or q to quit.\n"
        )
        self.assertEqual(actual, "b")

    def test_out_of_bounds_repeats(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["5", "2"])):
            with mock_stream() as stdout:
                actual = self._test()
        stdout = stdout.getvalue()
        self.assertEqual(
            stdout, "Please enter an index value between 1 and 3 " "or q to quit.\n"
        )
        self.assertEqual(actual, "b")

    def test_index_0_is_not_accepted(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["0", "2"])):
            with mock_stream() as stdout:
                actual = self._test()
        stdout = stdout.getvalue()
        self.assertEqual(
            stdout, "Please enter an index value between 1 and 3 " "or q to quit.\n"
        )
        self.assertEqual(actual, "b")

    def test_index_0_is_accepted_with_include_none(self):
        with mock.patch("builtins.input", lambda _: "0"):
            actual = self._test(True)
        self.assertIsNone(actual)

    def test_empty_input_prints_a_message_and_repeats(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["", "2"])):
            with mock_stream() as stdout:
                actual = self._test()
        stdout = stdout.getvalue()
        self.assertEqual(
            stdout, "Please enter an index value between 1 and 3 " "or q to quit.\n"
        )
        self.assertEqual(actual, "b")


class Confirm(unittest.TestCase):
    def test_y_is_true(self):
        with mock.patch("builtins.input", lambda _: "y"):
            self.assertTrue(interactive.confirm(""))

    def test_n_is_false(self):
        with mock.patch("builtins.input", lambda _: "n"):
            self.assertFalse(interactive.confirm(""))

    def test_Y_is_true(self):
        with mock.patch("builtins.input", lambda _: "Y"):
            self.assertTrue(interactive.confirm(""))

    def test_N_is_false(self):
        with mock.patch("builtins.input", lambda _: "N"):
            self.assertFalse(interactive.confirm(""))

    def test_empty_input_is_no(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["", "y"])):
            with mock_stream():
                self.assertFalse(interactive.confirm(""))

    def test_empty_input_can_be_forbidden_with_argument(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["", "y"])):
            with mock_stream():
                self.assertTrue(interactive.confirm("", False))



"""Tests for runtime type conversions"""

import datetime
import unittest

from khard.helpers.typing import (
    convert_to_vcard,
    list_to_string,
    string_to_date,
)


class ConvertToVcard(unittest.TestCase):
    def test_returns_strings(self):
        value = "some text"
        actual = convert_to_vcard("test", value, str)
        self.assertEqual(value, actual)

    def test_returns_lists(self):
        value = ["some", "text"]
        actual = convert_to_vcard("test", value, list)
        self.assertListEqual(value, actual)

    def test_fail_if_not_string(self):
        value = ["some", "text"]
        with self.assertRaises(ValueError):
            convert_to_vcard("test", value, str)

    def test_upgrades_string_to_list(self):
        value = "some text"
        actual = convert_to_vcard("test", value, list)
        self.assertListEqual([value], actual)

    def test_fails_if_string_lists_are_not_homogeneous(self):
        value = ["some", ["nested", "list"]]
        with self.assertRaises(ValueError):
            convert_to_vcard("test", value, list)

    def test_empty_list_items_are_filtered(self):
        value = ["some", "", "text", "", "more text"]
        actual = convert_to_vcard("test", value, list)
        self.assertListEqual(["some", "text", "more text"], actual)

    def test_strings_are_stripped(self):
        value = " some text "
        actual = convert_to_vcard("test", value, str)
        self.assertEqual("some text", actual)

    def test_strings_in_lists_are_stripped(self):
        value = [" some ", " text "]
        actual = convert_to_vcard("test", value, list)
        self.assertListEqual(["some", "text"], actual)


class ListToString(unittest.TestCase):
    def test_empty_list_returns_empty_string(self):
        the_list = []
        delimiter = " "
        expected = ""
        actual = list_to_string(the_list, delimiter)
        self.assertEqual(actual, expected)

    def test_simple_list(self):
        the_list = ["a", "bc", "def"]
        delimiter = " "
        expected = "a bc def"
        actual = list_to_string(the_list, delimiter)
        self.assertEqual(actual, expected)

    def test_simple_nested_list(self):
        the_list = ["a", "bc", ["x", "y", "z"], "def"]
        delimiter = " "
        expected = "a bc x y z def"
        actual = list_to_string(the_list, delimiter)
        self.assertEqual(actual, expected)

    def test_multi_level_nested_list(self):
        the_list = ["a", ["b", ["c", [[["x", "y"]]]]], "z"]
        delimiter = " "
        expected = "a b c x y z"
        actual = list_to_string(the_list, delimiter)
        self.assertEqual(actual, expected)

    def test_list_to_string_passes_through_other_objects(self):
        self.assertIs(list_to_string(None, "foo"), None)
        self.assertIs(list_to_string(42, "foo"), 42)
        self.assertIs(list_to_string("foo bar", "foo"), "foo bar")


class StringToDate(unittest.TestCase):
    date = datetime.datetime(year=1900, month=1, day=2)
    time = datetime.datetime(year=1900, month=1, day=2, hour=12, minute=42, second=17)
    zone = datetime.datetime(
        year=1900,
        month=1,
        day=2,
        hour=12,
        minute=42,
        second=17,
        tzinfo=datetime.timezone.utc,
    )

    def test_mmdd_format(self):
        string = "--0102"
        result = string_to_date(string)
        self.assertEqual(result, self.date)

    def test_mm_dd_format(self):
        string = "--01-02"
        result = string_to_date(string)
        self.assertEqual(result, self.date)

    def test_yyyymmdd_format(self):
        string = "19000102"
        result = string_to_date(string)
        self.assertEqual(result, self.date)

    def test_yyyy_mm_dd_format(self):
        string = "1900-01-02"
        result = string_to_date(string)
        self.assertEqual(result, self.date)

    def test_yyyymmddThhmmss_format(self):
        string = "19000102T124217"
        result = string_to_date(string)
        self.assertEqual(result, self.time)

    def test_yyyy_mm_ddThh_mm_ss_format(self):
        string = "1900-01-02T12:42:17"
        result = string_to_date(string)
        self.assertEqual(result, self.time)

    def test_yyyymmddThhmmssZ_format(self):
        string = "19000102T124217Z"
        result = string_to_date(string)
        self.assertEqual(result, self.time)

    def test_yyyy_mm_ddThh_mm_ssZ_format(self):
        string = "1900-01-02T12:42:17Z"
        result = string_to_date(string)
        self.assertEqual(result, self.time)

    def test_yyyymmddThhmmssz_format(self):
        string = "19000102T064217-06:00"
        result = string_to_date(string)
        self.assertEqual(result, self.zone)

    def test_yyyy_mm_ddThh_mm_ssz_format(self):
        string = "1900-01-02T06:42:17-06:00"
        result = string_to_date(string)
        self.assertEqual(result, self.zone)

    @unittest.expectedFailure
    def test_partial_date_with_february_29(self):
        # FIXME: this fails because 1900 is not a leap year, see
        # https://github.com/python/cpython/pull/116179 for more info.
        string = "--02-29"
        result = string_to_date(string)
        self.assertEqual(result, datetime.datetime(year=1900, month=2, day=29))



"""Tests for the custom YAML format."""

# pylint: disable=missing-docstring

import copy
import datetime
import unittest
from io import StringIO
from unittest import mock

from ruamel.yaml import YAML

import khard.helpers
from khard.contacts import Contact

from .helpers import TestYAMLEditable as create_test_card


def to_yaml(data):
    if "First name" not in data:
        data["First name"] = "Nobody"
    stream = StringIO()
    YAML().dump(data, stream)
    return stream.getvalue()


def parse_yaml(yaml: str = "") -> Contact:
    """Parse some yaml string into a Contact

    :param yaml: the yaml input string to parse
    :returns: the parsed Contact
    """
    return Contact.from_yaml(
        address_book=mock.Mock(path="foo-path"),
        yaml=yaml,
        supported_private_objects=[],
        version="3.0",
        localize_dates=False,
    )


class EmptyFieldsAndSpaces(unittest.TestCase):
    def test_empty_birthday_in_yaml_input(self):
        empty_birthday = "First name: foo\nBirthday:"
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(empty_birthday)
        self.assertIsNone(x.birthday)

    def test_only_spaces_in_birthday_in_yaml_input(self):
        spaces_birthday = "First name: foo\nBirthday:  "
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(spaces_birthday)
        self.assertIsNone(x.birthday)

    def test_empty_anniversary_in_yaml_input(self):
        empty_anniversary = "First name: foo\nAnniversary:"
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(empty_anniversary)
        self.assertIsNone(x.anniversary)

    def test_empty_organisation_in_yaml_input(self):
        empty_organisation = "First name: foo\nOrganisation:"
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(empty_organisation)
        self.assertListEqual(x.organisations, [])

    def test_empty_nickname_in_yaml_input(self):
        empty_nickname = "First name: foo\nNickname:"
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(empty_nickname)
        self.assertListEqual(x.nicknames, [])

    def test_empty_role_in_yaml_input(self):
        empty_role = "First name: foo\nRole:"
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(empty_role)
        self.assertListEqual(x.roles, [])

    def test_empty_title_in_yaml_input(self):
        empty_title = "First name: foo\nTitle:"
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(empty_title)
        self.assertListEqual(x.titles, [])

    def test_empty_categories_in_yaml_input(self):
        empty_categories = "First name: foo\nCategories:"
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(empty_categories)
        self.assertListEqual(x.categories, [])

    def test_empty_webpage_in_yaml_input(self):
        empty_webpage = "First name: foo\nWebpage:"
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(empty_webpage)
        self.assertListEqual(x.webpages, [])

    def test_empty_note_in_yaml_input(self):
        empty_note = "First name: foo\nNote:"
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(empty_note)
        self.assertListEqual(x.notes, [])

    def test_empty_kind_in_yaml_input(self):
        empty_kind = "First name: foo\nKind:"
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(empty_kind)
        self.assertEqual(x.kind, "individual")


class yaml_ablabel(unittest.TestCase):
    def test_ablabelled_url_in_yaml_input(self):
        ablabel_url = (
            "First name: foo\nWebpage:\n - http://example.com\n"
            " - github: https://github.com/scheibler/khard"
        )
        with mock.patch("khard.contacts.logger"):
            x = parse_yaml(ablabel_url)
        self.assertListEqual(
            x.webpages,
            ["http://example.com", {"github": "https://github.com/scheibler/khard"}],
        )


class UpdateVcardWithYamlUserInput(unittest.TestCase):
    _date = datetime.datetime(2000, 1, 1)
    _datetime = datetime.datetime(2013, 4, 2, 13, 14, 15)
    _no_year = datetime.datetime(1900, 1, 1)

    def test_update_org_simple(self):
        card = create_test_card()
        data = {"Organisation": "Foo"}
        data = to_yaml(data)
        card.update(data)
        self.assertListEqual(card.organisations, [["Foo"]])

    def test_update_org_multi(self):
        card = create_test_card()
        orgs = ["foo", "bar", "baz"]
        data = {"Organisation": orgs}
        data = to_yaml(data)
        card.update(data)
        self.assertListEqual(card.organisations, sorted([[x] for x in orgs]))

    def test_update_org_complex(self):
        card = create_test_card()
        org = ["org.", "dep.", "office"]
        data = {"Organisation": [org]}
        data = to_yaml(data)
        card.update(data)
        self.assertListEqual(card.organisations, [org])

    def test_update_categories_simple(self):
        card = create_test_card()
        data = {"Categories": "foo"}
        data = to_yaml(data)
        card.update(data)
        self.assertListEqual(card.categories, ["foo"])

    def test_update_categories_multi(self):
        card = create_test_card()
        cat = ["foo", "bar", "baz"]
        data = {"Categories": cat}
        data = to_yaml(data)
        card.update(data)
        self.assertListEqual(card.categories, cat)

    def test_update_bday_date(self):
        card = create_test_card()
        data = {"Birthday": "2000-01-01"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.birthday, self._date)

    def test_update_bday_without_year(self):
        card = create_test_card(version="4.0")
        data = {"Birthday": "--01-01"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.birthday, self._no_year)

    def test_update_bday_with_text(self):
        card = create_test_card(version="4.0")
        data = {"Birthday": "text= some day maybe"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.birthday, "some day maybe")

    def test_update_bday_with_date_and_time(self):
        card = create_test_card()
        data = {"Birthday": "2013-04-02T13:14:15"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.birthday, self._datetime)

    def test_update_anniversary(self):
        card = create_test_card()
        data = {"Anniversary": "2000-01-01"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.anniversary, self._date)

    def test_update_anniversary_without_year(self):
        card = create_test_card(version="4.0")
        data = {"Anniversary": "--01-01"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.anniversary, self._no_year)

    def test_update_anniversary_with_text(self):
        card = create_test_card(version="4.0")
        data = {"Anniversary": "text= some day maybe"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.anniversary, "some day maybe")

    def test_update_anniversary_with_date_and_time(self):
        card = create_test_card()
        data = {"Anniversary": "2013-04-02T13:14:15"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.anniversary, self._datetime)

    def test_update_name_simple(self):
        card = create_test_card()
        data = {"First name": "first", "Last name": "last"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.get_first_name_last_name(), "first last")

    def test_update_fn(self):
        card = create_test_card()
        fn = "me myself and i"
        data = {"Formatted name": fn}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.formatted_name, fn)

    def test_update_kind(self):
        card = create_test_card(version="4.0")
        self.assertEqual(card.kind, "individual")
        data = {"Kind": "org"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.kind, "org")

    def test_update_kind_on_3_0_card(self):
        card = create_test_card()
        self.assertEqual(card.kind, "individual")
        data = {"Kind": "org"}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.kind, "org")

    def test_update_with_bad_kind_field_fails(self):
        card = create_test_card(version="4.0")
        self.assertEqual(card.kind, "individual")
        data = to_yaml({"Kind": "foo"})
        with self.assertRaises(ValueError):
            card.update(data)
        self.assertEqual(card.kind, "individual")

    def test_parse_field(self):
        """Test round-trip of a field to/from YAML"""
        card = create_test_card()
        data = "First name: Nobody\n"
        data += "\n".join(khard.helpers.convert_to_yaml("Note", "foobar", 0, 5, True))
        card.update(data)
        self.assertListEqual(card.notes, ["foobar"])

    def test_parse_field_with_colon(self):
        """Test round-trip of a field containing ': ' to/from YAML"""
        card = create_test_card()
        data = "First name: Nobody\n"
        data += "\n".join(khard.helpers.convert_to_yaml("Note", "foo: bar", 0, 5, True))
        card.update(data)
        self.assertListEqual(card.notes, ["foo: bar"])

    def test_vcard_round_trip(self):
        """Test a VCARD can be converted to YAML and back unchanged"""
        card = create_test_card()
        card._add_organisation("ACME, Inc")
        card._add_note("foo: bar")
        card2 = copy.copy(card)
        yaml = card.to_yaml()
        card.update(yaml)
        self.assertEqual(card.vcard.serialize(), card2.vcard.serialize())



"""Tests for the helpers module."""

# pylint: disable=missing-docstring

import unittest

from khard import helpers


class ConvertToYAML(unittest.TestCase):
    def test_colon_handling(self):
        result = helpers.convert_to_yaml("Note", "foo: bar", 0, 5, True)
        self.assertListEqual(result, ["Note : |\n    foo: bar"])

    def test_none_values_produce_no_output(self):
        result = helpers.convert_to_yaml("Note", None, 0, 5, True)
        self.assertListEqual(result, [])

    def test_empty_strings_produce_empty_values(self):
        result = helpers.convert_to_yaml("Note", "", 0, 5, True)
        self.assertListEqual(result, ["Note : "])

    def test_preparing_multiple_addresses_with_same_label_for_yaml_conversion_returns_all_entries(
        self,
    ):
        input = {
            "home": [
                {"street": "street 1", "city": "city1", "code": "zip1", "country": ""},
                {"street": "street 2", "city": "city2", "code": "zip2", "country": ""},
            ]
        }
        expected = [
            {"Street": "street 1", "City": "city1", "Code": "zip1", "Country": None},
            {"Street": "street 2", "City": "city2", "Code": "zip2", "Country": None},
        ]
        actual = helpers.yaml_addresses(input, ["Street", "Code", "City", "Country"])
        self.assertEqual(expected, actual["home"])

    def test_preparing_single_address_for_yaml_conversion_returns_dict_not_list(self):
        input = {
            "home": [{"street": "street", "city": "city", "code": "zip", "country": ""}]
        }
        expected = {"Street": "street", "City": "city", "Code": "zip", "Country": None}
        actual = helpers.yaml_addresses(input, ["Street", "Code", "City", "Country"])
        self.assertEqual(expected, actual["home"])


if __name__ == "__main__":
    unittest.main()
